from collections.abc import Iterable
from .models import Pet, Person, Driver, Car


def test_relation(nrm):
    org = Car(
        name      = 'Testing Inc',
    ).save()

    user = Driver(
        name      = 'Sam',
    ).save()
    user.proxy.cars.set([org])

    assert user.proxy.cars.count() == 1

    assert user.cars[0].id == org.id
    assert nrm.redis.sismember('driver:{}:srel_cars'.format(user.id), org.id)

    same = Driver.get(user.id)
    same.proxy.cars.fill()

    assert same.cars[0].id == org.id
    assert same.to_json(include=['cars'])['cars'][0]['name'] == 'Testing Inc'

def test_delete_with_related(nrm):
    doggo = Pet(name='doggo').save()
    catto = Pet(name='catto').save()

    owner = Person(
        name = 'John',
    ).save()
    owner.proxy.pets.set([doggo, catto])

    assert doggo in owner.proxy.pets
    assert catto in owner.proxy.pets

    owner.delete()

    assert Pet.get(doggo.id) is None
    assert Pet.get(catto.id) is None

def test_many_to_many_relationship(nrm):
    u1 = Driver(name='u1').save()
    u2 = Driver(name='u2').save()
    o1 = Car(name='o1').save()
    o2 = Car(name='o2').save()

    # test adding a relationship creates the inverse
    u1.proxy.cars.set([o1, o2])

    o1.proxy.drivers.fill()
    o2.proxy.drivers.fill()

    assert type(o1.drivers) == list
    assert type(o2.drivers) == list

    assert len(o1.drivers) == 1
    assert len(o2.drivers) == 1

    assert o1.drivers[0].name == 'u1'
    assert o2.drivers[0].name == 'u1'

    u2.proxy.cars.set([o2])

    o1.proxy.drivers.fill()
    o2.proxy.drivers.fill()

    assert type(o1.drivers) == list
    assert type(o2.drivers) == list

    assert len(o1.drivers) == 1
    assert len(o2.drivers) == 2

    o2.drivers.sort(key=lambda x:x.name)

    assert o1.drivers[0].name == 'u1'
    assert o2.drivers[0].name == 'u1'
    assert o2.drivers[1].name == 'u2'

    # test deleting an object deletes the relationship in the related

    u1.delete()

    o1.proxy.drivers.fill()
    o2.proxy.drivers.fill()

    assert type(o1.drivers) == list
    assert type(o2.drivers) == list

    assert len(o1.drivers) == 0
    assert len(o2.drivers) == 1

def test_querying_related(nrm):
    o1 = Car(name='o1').save()
    u1 = Driver(name='u1').save()
    u2 = Driver(name='u2').save()
    u1.proxy.cars.set([o1])

    assert u1 in o1.proxy.drivers
    assert u2 not in o1.proxy.drivers

def test_can_filter_related(nrm):
    p = Person(name='Juan').save()

    pets = [
        Pet(name='bc').save(), # 0
        Pet(name='bd').save(), # 1
        Pet(name='cd').save(), # 2
    ]

    p.proxy.pets.set(pets)

    assert isinstance(p.proxy.pets.q().filter(name__startswith='pa'), Iterable)

    res = list(map(
        lambda x:x.id,
        p.proxy.pets.q().filter(name__startswith='b', name__endswith='d')
    ))

    assert res == [pets[1].id]

    res = list(map(
        lambda x:x.id,
        p.proxy.pets.q().filter(name__startswith='b').filter(name__endswith='d')
    ))

    assert res == [pets[1].id]

def test_foreign_key(nrm):
    owner = Person(name='John').save()
    pet = Pet(name='doggo').save()

    pet.proxy.owner.set(owner)
    owner.proxy.pets.fill()

    assert pet.owner is not None
    assert pet.owner.id == owner.id

    assert type(owner.pets) == list
    assert len(owner.pets) == 1
    assert owner.pets[0].id == pet.id

    pet.delete()

    pet.proxy.owner.fill()
    owner.proxy.pets.fill()

    assert type(owner.pets) == list
    assert len(owner.pets) == 0

def test_foreign_key_inverse(nrm):
    pet = Pet(name='doggo').save()
    owner = Person(name='John').save()
    owner.proxy.pets.set([pet])

    pet.proxy.owner.fill()
    owner.proxy.pets.fill()

    assert pet.owner is not None
    assert pet.owner.id == owner.id

    assert type(owner.pets) == list
    assert len(owner.pets) == 1
    assert owner.pets[0].id == pet.id

def test_delete_relation(nrm):
    c1 = Car().save()
    c2 = Car().save()

    d1 = Driver().save()

    d1.proxy.cars.set([c1, c2])

    assert c1 in d1.proxy.cars
    assert c2 in d1.proxy.cars
    assert d1 in c1.proxy.drivers
    assert d1 in c2.proxy.drivers

    d1.proxy.cars.remove(c1)

    assert c1 not in d1.proxy.cars
    assert c2 in d1.proxy.cars
    assert d1 not in c1.proxy.drivers
    assert d1 in c2.proxy.drivers

def test_get_relation(nrm):
    c1 = Car().save()
    c2 = Car().save()

    d1 = Driver().save()

    d1.proxy.cars.set([c1, c2])

    cars = d1.proxy.cars.get()

    orig = sorted([c1, c2], key=lambda c:c.id)
    cars = sorted(cars, key=lambda c:c.id)

    assert orig == cars

def test_get_foreignid_relation(nrm):
    pet = Pet().save()
    owner = Person().save()
    owner.proxy.pets.add(pet)

    assert pet.proxy.owner.get().id == owner.id

    owner.proxy.pets.remove(pet)

    assert pet.proxy.owner.get() is None



from coralillo import Model, Form, BoundedModel, Engine, fields
from coralillo.validation import validation_rule
from coralillo.errors import InvalidFieldError
import sys, inspect

# for model testing
class Table(Model):
    name = fields.Text()

class Ship(Model):
    name = fields.Text()
    code = fields.Text(index=True)

class Tenanted(BoundedModel):
    name = fields.Text()

    @classmethod
    def prefix(cls):
        return 'testing'

class SideWalk(Model):
    name = fields.Text()

class House(Model):
    number = fields.Integer(required=False)

# for events test
class Something(Model):
    name = fields.Text()
    notify = True

# for fields test
class Subscription(Model):
    key_name = fields.TreeIndex()

class User(Model):
    password = fields.Hash()

class Truck(Model):
    last_position = fields.Location()

# for form testing
class ShipForm(Form):
    name = fields.Text()
    code = fields.Text()

# for testing permissions
class Bunny(BoundedModel):
    name = fields.Text()

    @classmethod
    def prefix(cls):
        return 'bound'

# for testing validations
class MyForm(Form):
    field1 = fields.Text()
    field2 = fields.Text()

    @validation_rule
    def enforce_fields(data):
        if (data.field1 is None and data.field2 is None) or \
            (data.field1 is not None and data.field2 is not None):
            raise InvalidFieldError(field='field1')

class MyModel(Model):
    field1 = fields.Text(index=True, required=False, private=True)

# For testing relationships
class Pet(Model):
    name = fields.Text()
    owner = fields.ForeignIdRelation('coralillo.tests.models.Person', inverse='pets')

class Person(Model):
    name = fields.Text()
    pets = fields.SetRelation(Pet, on_delete='cascade', inverse='owner')

# For many to many relations
class Driver(Model):
    name = fields.Text()
    cars = fields.SetRelation('coralillo.tests.models.Car', inverse='drivers')

class Car(Model):
    name = fields.Text()
    drivers = fields.SetRelation(Driver, inverse='cars')

class Employee(Model):
    name = fields.Text()
    last_name = fields.Text()
    office = fields.ForeignIdRelation('coralillo.tests.models.Office', inverse='employees')

class Office(Model):
    name = fields.Text()
    address = fields.Text()
    employees = fields.SetRelation(Employee, inverse='office')

# for testing nested embedding
class A(Model):
    attr = fields.Text()
    bs = fields.SetRelation('coralillo.tests.models.B', inverse='a')

class B(Model):
    attr = fields.Text()
    a = fields.ForeignIdRelation(A, inverse='bs')
    cs = fields.SetRelation('coralillo.tests.models.C', inverse='b')

class C(Model):
    attr = fields.Text()
    b = fields.ForeignIdRelation(A, inverse='cs')

def bound_models(eng):
    for name, cls in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(cls):
            if issubclass(cls, Form):
                cls.set_engine(eng)



from coralillo import Engine, Model, datamodel
from coralillo.fields import *
from coralillo.hashing import check_password, make_password
from datetime import datetime
import time
import os
import pytest

from .models import Truck, User, Subscription

os.environ['TZ'] = 'UTC'
time.tzset()

def test_field_text():
    field = Text(name='field', default='pollo')

    assert field.recover({'field': 'foo'}, 'field') == 'foo'
    assert field.prepare('foo') == 'foo'
    assert field.validate(None, 'field') == 'pollo'
    assert field.recover({'field': ''}, 'field') == ''
    assert field.recover({'field': None}, 'field') == None
    assert field.recover({'field': 'None'}, 'field') == None

def test_field_text_validate():
    field = Text(name='field')

    with pytest.raises(MissingFieldError):
        field.validate('', None)

def test_field_hash():
    field = Hash(name='field')

    assert field.recover({'field': 'foo'}, 'field') == 'foo'
    assert field.validate('foo', 'field') != 'foo'
    assert field.recover({'field': None}, 'field') == None
    assert field.recover({'field': 'None'}, 'field') == None

def test_field_bool():
    field = Bool(name='field')

    assert field.recover({'field': 'True'}, 'field') == True
    assert field.recover({'field': 'False'}, 'field') == False
    assert field.prepare(True) == 'True'
    assert field.prepare(False) == 'False'
    assert field.recover({'field': None}, 'field') == None
    assert field.recover({'field': 'None'}, 'field') == None

    assert field.validate('true', 'field') == True
    assert field.validate('false', 'field') == False
    assert field.validate('1', 'field') == True
    assert field.validate('0', 'field') == False
    assert field.validate('', 'field') == False

def test_field_int():
    field = Integer(name='field')

    assert field.recover({'field': '35'}, 'field') == 35
    assert field.recover({'field': '0'}, 'field') == 0
    assert field.recover({'field': ''}, 'field') == None
    assert field.recover({'field': None}, 'field') == None
    assert field.recover({'field': 'None'}, 'field') == None

    assert field.prepare(35) == '35'
    assert field.prepare(0) == '0'

    assert field.validate(0, 'field') == 0
    assert field.validate('0', 'field') == 0

    with pytest.raises(MissingFieldError):
        field.validate('', 'field')

    with pytest.raises(InvalidFieldError):
        field.validate('pollo', 'field')

def test_field_int_no_required():
    field = Integer(name='field', required=False)

    assert field.validate(None, 'field') == None
    assert field.validate('', 'field') == None

def test_field_float():
    field = Float(name='field')

    assert field.recover({'field': '3.14'}, 'field') == 3.14
    assert field.recover({'field': '0'}, 'field') == 0.0
    assert field.recover({'field': ''}, 'field') == None
    assert field.recover({'field': None}, 'field') == None
    assert field.recover({'field': 'None'}, 'field') == None

    assert field.prepare(3.14) == '3.14'
    assert field.prepare(0.0) == '0.0'

    assert field.validate(0, 'field') == 0
    assert field.validate('0', 'field') == 0

def test_field_date():
    field = Datetime(name='field')

    assert field.recover({'field': '1499794899'}, 'field') == datetime(2017, 7, 11, 17, 41, 39)
    assert field.recover({'field': ''}, 'field') == None
    assert field.recover({'field': None}, 'field') == None
    assert field.recover({'field': 'None'}, 'field') == None
    assert field.prepare(datetime(2017, 7, 11, 17, 41, 39)) == '1499794899'

def test_field_location(nrm):
    class MyModel(Model):
        field = Location()

        class Meta:
            engine = nrm

    obj = MyModel(field = datamodel.Location(-103.3590, 20.7240)).save()

    assert nrm.redis.type('my_model:geo_field') == b'zset'
    assert MyModel.get(obj.id).field == datamodel.Location(-103.3590, 20.7240)

def test_field_location_validate(nrm):
    class MyModel(Model):
        field = Location()

        class Meta:
            engine = nrm

    item = MyModel.validate(field='-100,20').save()
    assert item.field == datamodel.Location(-100, 20)

    item_rec = MyModel.get(item.id)
    assert item_rec.field == datamodel.Location(-100, 20)

def test_field_location_none(nrm):
    class MyModel(Model):
        field = Location(required=False)

        class Meta:
            engine = nrm

    item = MyModel.validate().save()

    item_rec = MyModel.get(item.id)

    assert item_rec.field is None

def test_password_check():
    user = User(
        password  = make_password('123456'),
    ).save()

    assert check_password('123456', user.password)
    assert user.password != '123456'

def test_empty_field_dict(nrm):
    class Dummy(Model):
        dynamic = Dict()

        class Meta:
            engine = nrm

    a = Dummy().save()
    loaded_a = Dummy.get(a.id)

    assert loaded_a.dynamic == {}

def test_field_dict(nrm):
    class Dummy(Model):
        name           = Text()
        dynamic        = Dict()

        class Meta:
            engine = nrm

    # Create and save models with a dict field
    a = Dummy(
        name = 'dummy',
        dynamic = {
            '1': 'one',
        },
    ).save()

    b = Dummy(
        name = 'dummy',
        dynamic = {
            '2': 'two',
        },
    ).save()

    c = Dummy(
        name = 'dummy',
        dynamic = {
            '1': 'one',
            '2': 'two',
        },
    ).save()

    # override dict
    a.dynamic = {
        '3': 'thre',
    }

    a.save()

    loaded_a = Dummy.get(a.id)
    loaded_b = Dummy.get(b.id)
    loaded_c = Dummy.get(c.id)

    assert a.dynamic == loaded_a.dynamic
    assert b.dynamic == loaded_b.dynamic
    assert c.dynamic == loaded_c.dynamic

def test_field_position_delete():
    c1 = Truck(last_position=datamodel.Location(-90, 21)).save()
    c2 = Truck(last_position=datamodel.Location(-90, 22)).save()

    assert Truck.get(c1.id).last_position == datamodel.Location(-90, 21)
    assert Truck.get(c2.id).last_position == datamodel.Location(-90, 22)

    c1.delete()

    assert Truck.get(c1.id) is None

    assert Truck.get(c2.id).last_position == datamodel.Location(-90, 22)

def test_tree_index():
    s4 = Subscription(key_name='a:b:c:d').save()
    s1 = Subscription(key_name='a:b:c').save()
    s2 = Subscription(key_name='a:b').save()
    s3 = Subscription(key_name='a').save()

    assert Subscription.tree_match('key_name', 'a:b:c') == [s1, s2, s3]

    s3.delete()
    assert Subscription.tree_match('key_name', 'a:b:c') == [s1, s2]

    s2.delete()
    assert Subscription.tree_match('key_name', 'a:b:c') == [s1]

    s1.delete()
    assert Subscription.tree_match('key_name', 'a:b:c') == []
